#define CORE_NAME "eth_rma1"

#include <autoconf.h>
#if defined(CONFIG_ETHERNET_DEBUG)
#define DEBUG
#define TRACE_DEBUG
#endif
#include <hermit.h>
#include <herrno.h>
#include <io.h>
#include <arch/pinmux.h>
#include <string.h>
#include <mac.h>
#include <net/net.h>
#include <net/eth.h>
#include <net/eth_util.h>
#include <delay.h>

#define xPHYRST	PIN_FMSOSLD

static void gpio_output(u32 pin, u8 level)
{
	if (level)
		write32(PIN_GPIODSR(pin), 1 << PIN_GPIO32(pin));
	else
		write32(PIN_GPIODCR(pin), 1 << PIN_GPIO32(pin));
}

#define GETHER_ARSTR		0xe9a01800

#define GETHER_EDSR		0xe9a00000
#define GETHER_TDLAR		0xe9a00010
#define GETHER_TDFAR		0xe9a00014
#define GETHER_TDFXR		0xe9a00018
#define GETHER_TDFFR		0xe9a0001c
#define GETHER_RDLAR		0xe9a00030
#define GETHER_RDFAR		0xe9a00034
#define GETHER_RDFXR		0xe9a00038
#define GETHER_RDFFR		0xe9a0003c

#define GETHER_EDMR		0xe9a00400
#define GETHER_EDTRR		0xe9a00408
#define GETHER_EDRRR		0xe9a00410
#define GETHER_EESR		0xe9a00428
#define GETHER_EESIPR		0xe9a00430

#define GETHER_TRSCER		0xe9a00438
#define GETHER_TFTR		0xe9a00448

#define GETHER_FDR		0xe9a00450
#define GETHER_RMCR		0xe9a00458
#define GETHER_RPADIR		0xe9a00460
#define GETHER_FCFTR		0xe9a00468

#define GETHER_APR		0xe9a00554
#define GETHER_MPR		0xe9a00558
#define GETHER_TPAUSER		0xe9a00564

#define GETHER_ECMR		0xe9a00500
#define GETHER_RFLR		0xe9a00508
#define GETHER_ECSR		0xe9a00510
#define GETHER_ECSIPR		0xe9a00518

#define GETHER_PIR0		0xe9a00520
#define PIR_MDC	(1 << 0)
#define PIR_MMD (1 << 1)
#define PIR_MDO (1 << 2)
#define PIR_MDI (1 << 3)
#define GETHER_PIPR		0xe9a0052c

#define GETHER_GECMR		0xe9a005b0
#define GETHER_MAHR		0xe9a005c0
#define GETHER_MALR		0xe9a005c8


#define GETHER_RMII_MII		0xe9a00790

#define GETHER_TSU_FWSLC	0xe9a01838

struct tx_desc {
	union {
		u32 v;
		struct {
			u32 tfs:12,
			    rsv:14,
			    twbi:1,
			    tfe:1,
			    tfp:2,
			    tdle:1,
			    tact:1;
		} s;
	} td0;
	union {
		u32 v;
		struct {
			u32 rsv:16,
			    tdl:16;
		} s;
	} td1;
	union {
		u32 v;
		struct {
			u32 tba;
		} s;
	} td2;
	u32 padding;
};

struct rx_desc {
	union {
		u32 v;
		struct {
			u32 rcs:16,
			    rfs:10,
			    rcse:1,
			    rfe:1,
			    rfp:2,
			    rdle:1,
			    ract:1;
		} s;
	} rd0;
	union {
		u32 v;
		struct {
			u32 rdl:16,
			    rbl:16;
		} s;
	} rd1;
	union {
		u32 v;
		struct {
			u32 rba;
		} s;
	} rd2;
	u32 padding;
};

#define nr_tx_desc (4)
#define nr_rx_desc (8)
#define desc_buf_size (2048)
static volatile struct tx_desc *tx_desc = (struct tx_desc *)0x40b00000;
static volatile struct rx_desc *rx_desc = (struct rx_desc *)0x40b20000;
static volatile u8 *tx_buf = (u8 *)0x40b40000;
static volatile u8 *rx_buf = (u8 *)0x40b80000;
static volatile struct tx_desc *tx_desc_cur = (struct tx_desc *)0x40b00000;
static volatile struct rx_desc *rx_desc_cur = (struct rx_desc *)0x40b20000;

static void dump(u32 addr, int len)
{
#if defined(DEBUG_DUMP)
	u8 *ptr = (u8 *)addr;
	int i;

	for (i=0; i<len; i++) {
		dev_dbg_r("%b ", *ptr++);
		if (i%16==15)
			dev_dbg_r("\n");
	}
	if (i%16!=16)
		dev_dbg_r("\n");
#endif
}

static u8 __mii_read_bit(void)
{
	u32 pir;
	write32(GETHER_PIR0, PIR_MDC);
	udelay(1);
	write32(GETHER_PIR0, 0);
	udelay(2);
	pir = read32(GETHER_PIR0);
	write32(GETHER_PIR0, PIR_MDC);
	udelay(1);

	return pir & PIR_MDI;
}

static void __mii_write_bit(u32 level)
{
	u32 mdo = level ? PIR_MDO : 0;

	write32(GETHER_PIR0, mdo | PIR_MMD | PIR_MDC);
	udelay(1);
	write32(GETHER_PIR0, mdo | PIR_MMD);
	udelay(2);
	write32(GETHER_PIR0, mdo | PIR_MMD | PIR_MDC);
	udelay(1);
}

static u16 mdio_read(struct net_device *dev, addr_t addr, addr_t reg, u16 *val)
{
	u16 _val = 0;
	int i;

	/* Preamble */
	for (i=0; i<32; i++)
		__mii_write_bit(1);

	/* Start of Frame */
	__mii_write_bit(0);
	__mii_write_bit(1);

	/* OP Code: Read */
	__mii_write_bit(1);
	__mii_write_bit(0);

	/* PHY Address */
	for (i=4; i>=0; i--)
		__mii_write_bit(addr & (1 << i));

	/* Register Address */
	for (i=4; i>=0; i--)
		__mii_write_bit(reg & (1 << i));

	/* Turn Around */
	__mii_read_bit();
	__mii_read_bit();

	/* Data */
	for (i=15; i>=0; i--)
		_val |= ((__mii_read_bit() ? 1 : 0) << i);

	if (val)
		*val = _val;

	return 0;
}

static int mdio_write(struct net_device *dev, addr_t addr, addr_t reg, u16 val)
{
	int i;

	/* Preamble */
	for (i=0; i<32; i++)
		__mii_write_bit(1);

	/* Start of Frame */
	__mii_write_bit(0);
	__mii_write_bit(1);

	/* OP Code: Write */
	__mii_write_bit(0);
	__mii_write_bit(1);

	/* PHY Address */
	for (i=4; i>=0; i--)
		__mii_write_bit(addr & (1 << i));

	/* Register Address */
	for (i=4; i>=0; i--)
		__mii_write_bit(reg & (1 << i));

	/* Turn Around */
	__mii_write_bit(0);
	__mii_write_bit(0);

	/* Data */
	for (i=15; i>=0; i--)
		__mii_write_bit(val & (1 << i));

	return 0;
}

static void rma1_enable_phy(struct net_device *dev)
{
	trace();
}

static void rma1_disable_phy(struct net_device *dev)
{
	trace();
}

static void rma1_phy_reset(void)
{
	gpio_output(xPHYRST, 0);
	udelay(100);
	gpio_output(xPHYRST, 1);
}

static int rma1_phy_init(struct net_device *dev)
{
	int timeout;
	u16 val;
	u32 id;
	u16 id1, id2;
	trace();

	rma1_phy_reset();
	mdio_read(dev, 0, 2, &id1);
	mdio_read(dev, 0, 3, &id2);
	id = (id1 << 16 | id2);
	dev_dbg("PHYID = %p, REV = %d\n", id >> 4, id & 0x0f);

	/* wait auto-negotiation complete */
	mdio_read(dev, 0x00, 0x00, &val);
	mdio_write(dev, 0x00, 0x00, 0x1200);
	mdio_read(dev, 0x00, 0x00, &val);

	timeout = 5000;
	while (--timeout) {
		mdio_read(dev, 0x00, 0x01, &val);
		if (val & 0x0020)
			break;
		mdelay(1);
		dev_dbg("Autonegotiation Satus: %p\n", val);
	}
	mdio_read(dev, 0x00, 0x01, &val);

	if (val & 0x0020) {
		dev_dbg("Autonegotiation Complete! %p\n", val);

	} else {
	}
	if (!(val & 0x04)) {
		dev_dbg_r("Link is down!");
		return -H_EIO;
	}
	dev_dbg_r("Link is up!\n");

	mdio_read(dev, 0x00, 31, &val);
	dev_dbg("%s, %s-duplex\n",
		((val & 0xc) == 0x8) ? "100BASE-TX" :
		 (((val & 0xc) == 0x4) ? "10BASE-T" : "unknown speed"),
		 (val & 0x10) ? "Full" : "Half");

	return 0;
}

static void rma1_eth_reset(void)
{
	trace();

	cpg_clr_mstpcr(309);
	write32(GETHER_RMII_MII, 1); /* MII */
}

static void rma1_setup_tx_desc(struct net_device *dev)
{
	int i;
	trace();

	for(i=0; i<nr_tx_desc; i++) {
		memset((void *)&tx_desc[i], 0, sizeof(struct tx_desc));
		tx_desc[i].td0.s.tact = 0;
		tx_desc[i].td1.s.tdl = desc_buf_size;
		tx_desc[i].td2.s.tba = (u32)(tx_buf + (desc_buf_size * i));
	}
	tx_desc[nr_tx_desc - 1].td0.s.tdle = 1;

	write32(GETHER_TDLAR, (u32)tx_desc);
	write32(GETHER_TDFAR, (u32)tx_desc);
	write32(GETHER_TDFXR, (u32)&tx_desc[nr_tx_desc - 1]);
	write32(GETHER_TDFFR, 1);
	tx_desc_cur = tx_desc;
}

static void rma1_setup_rx_desc(struct net_device *dev)
{
	int i;
	trace();

	for (i=0; i<nr_rx_desc; i++) {
		memset((void *)&rx_desc[i], 0, sizeof(struct rx_desc));
		rx_desc[i].rd0.s.ract = 1;
		rx_desc[i].rd1.s.rbl = desc_buf_size;
		rx_desc[i].rd2.s.rba = (u32)(rx_buf + (desc_buf_size * i));
	}
	rx_desc[nr_rx_desc - 1].rd0.s.rdle = 1;

	write32(GETHER_RDLAR, (u32)rx_desc);
	write32(GETHER_RDFAR, (u32)rx_desc);
	write32(GETHER_RDFXR, (u32)&rx_desc[nr_rx_desc - 1]);
	write32(GETHER_RDFFR, 1);
	rx_desc_cur = rx_desc;
}

static int rma1_send(struct net_device *dev, struct net_packet *pkt)
{
	volatile struct tx_desc *desc;
	int len;
	u8 *buf;
	int to = 1000000;

	trace();

	if (tx_desc_cur == &tx_desc[nr_tx_desc])
		tx_desc_cur = tx_desc;

	desc = tx_desc_cur;

	buf = (u8 *)desc->td2.s.tba;
	memcpy(buf, pkt->eth, ETH_FRAME_LEN);
	len = ETH_FRAME_LEN;
	memcpy(buf + len, pkt->f2.any, pkt->f2len);
	len += pkt->f2len;
	memcpy(buf + len, pkt->f3.any, pkt->f3len);
	len += pkt->f3len;
	memcpy(buf + len, pkt->f4, pkt->f4len);
	len += pkt->f4len;

	if (len < 60)
		len = 60;


	desc->td0.s.tfs = 0;
	desc->td0.s.tfe = 0;
	desc->td0.s.tfp = 3;
	desc->td1.s.tdl = len;
	desc->td0.s.tact = 1;

	write32(GETHER_EESR, 0x20200000);

	if (!(read32(GETHER_EDTRR) & 0x3))
		write32(GETHER_EDTRR, 0x3);

	while (to--) {
		if (!desc->td0.s.tact)
			break;
		udelay(1);
	}
	while (to--) {
		if ((read32(GETHER_EESR) & 0x20200000) == 0x20200000)
			break;
		udelay(1);
	}

	if (desc->td0.s.tact)
		dev_dbg("Tx timeout\n");

	tx_desc_cur++;

	return 0;
}

static int rma1_recv(struct net_device *dev, struct net_packet *pkt,
		     int *timeout)
{
	volatile struct rx_desc *desc;
	int to = 3000000;
	trace();

retry:
	if (rx_desc_cur == &rx_desc[nr_rx_desc])
		rx_desc_cur = rx_desc;

	desc = rx_desc_cur;

	while (to--) {
		if (!desc->rd0.s.ract)
			break;
		udelay(1);
	}

	if (!desc->rd0.s.ract) {
		if (desc->rd0.s.rfe) {
			desc->rd0.s.rcs = 0;
			desc->rd0.s.rfs = 0;
			desc->rd0.s.rcse = 0;
			desc->rd0.s.rfe = 0;
			desc->rd0.s.rfp = 0;
			desc->rd1.s.rdl = 0;
			desc->rd0.s.ract = 1;

			if (read32(GETHER_EDRRR) != 1)
				write32(GETHER_EDRRR, 1);

			rx_desc_cur++;
			goto retry;
		}

		dump(desc->rd2.s.rba, desc->rd1.s.rdl);
		memcpy(pkt->raw_data_addr, (void *)desc->rd2.s.rba,
		       desc->rd1.s.rdl);
		pkt->raw_data_len = desc->rd1.s.rdl;

		desc->rd0.s.rcs = 0;
		desc->rd0.s.rfs = 0;
		desc->rd0.s.rcse = 0;
		desc->rd0.s.rfe = 0;
		desc->rd0.s.rfp = 0;
		desc->rd1.s.rdl = 0;
		desc->rd0.s.ract = 1;

		if (read32(GETHER_EDRRR) != 1)
			write32(GETHER_EDRRR, 1);

		rx_desc_cur++;

		return 0;
	}
	dev_dbg("Not Received!\n");

	return -1;
}

static void rma1_set_mac(struct net_device *dev, unsigned char *mac)
{
	u32 mac_val;
	mac_val = (mac[0] << 24 |
		   mac[1] << 16 |
		   mac[2] << 8 |
		   mac[3] << 0);
	write32(GETHER_MAHR, mac_val);
	mac_val = (mac[4] << 8 |
		   mac[5] << 0);
	write32(GETHER_MALR, mac_val);
}

static int rma1_eth_init(struct net_device *dev)
{
	u16 val;
	int ret;

	trace();

	ret = PIN_SDHID0_2;
	ret = MUX(3, 0, 1, PULL_OFF);
	pinmux(PIN_SDHID0_2,	MUX(3, 0, 1, PULL_OFF)); /* CRS */
	pinmux(PIN_SDHID1_2,	MUX(3, 1, 0, PULL_OFF)); /* MDC */
	pinmux(PIN_SDHID2_2,	MUX(3, 1, 1, PULL_OFF)); /* MDIO */
	pinmux(PIN_SDHICLK_2,	MUX(3, 1, 0, PULL_OFF)); /* TX_ER */
	pinmux(PIN_SDHICMD_2,	MUX(3, 0, 1, PULL_OFF)); /* RX_ER */
	pinmux(PIN_SCIFA_SCK_1,	MUX(3, 0, 1, PULL_OFF)); /* ERXD0 */
	pinmux(PIN_SCIFB_RTS_N,	MUX(3, 0, 1, PULL_OFF)); /* ERXD1 */
	pinmux(PIN_SCIFB_CTS_N,	MUX(3, 0, 1, PULL_OFF)); /* ERXD2 */
	pinmux(PIN_SCIFA_SCK_0,	MUX(3, 0, 1, PULL_OFF)); /* ERXD3 */
	pinmux(PIN_DACK_1_N,	MUX(3, 0, 1, PULL_OFF)); /* TX_CLK */
	pinmux(PIN_DREQ_1_N,	MUX(3, 1, 0, PULL_OFF)); /* TX_EN */
	pinmux(PIN_D16,		MUX(3, 1, 0, PULL_OFF)); /* ETXD0 */
	pinmux(PIN_D17,		MUX(3, 1, 0, PULL_OFF)); /* ETXD1 */
	pinmux(PIN_D18,		MUX(3, 1, 0, PULL_OFF)); /* ETXD2 */
	pinmux(PIN_D19,		MUX(3, 1, 0, PULL_OFF)); /* ETXD3 */
	pinmux(PIN_D25,		MUX(3, 0, 1, PULL_UP)); /* PHY_INT */
	pinmux(PIN_D26,		MUX(3, 0, 1, PULL_OFF)); /* COL */
	pinmux(PIN_D28,		MUX(3, 0, 1, PULL_OFF)); /* RX_DV */
	pinmux(PIN_A26,		MUX(3, 0, 1, PULL_OFF)); /* RX_CLK */
	pinmux(PIN_FMSOSLD,	MUX(0, 1, 0, PULL_OFF)); /* PHYRST */

	if (!is_valid_mac(dev->eth_mac))
		get_mac_address(dev->eth_mac);

	enable_phy_module(dev);

	rma1_eth_reset();

	ret = rma1_phy_init(dev);
	if (ret) {
		dev_info_r("Link is down\n");
		return ret;
	}

	/* setup */
	write32(GETHER_EDSR, 3);
	write32(GETHER_EDMR, 0x03);
	while (read32(GETHER_EDMR) & 0x3);

	/* setup rx_desc */
	rma1_setup_tx_desc(dev);
	rma1_setup_rx_desc(dev);

	/* setup dmac */
	write32(GETHER_EDMR, 0x40);
	write32(GETHER_EESIPR, 0);
	write32(GETHER_TRSCER, 0);
	write32(GETHER_TFTR, 0);
	write32(GETHER_FDR, 0x00000707);
	write32(GETHER_RMCR, 0x00000001);
	write32(GETHER_RPADIR, 0);
	write32(GETHER_FCFTR, 0x001700ff);

	/* setup emac */
	write32(GETHER_ECSIPR, 0);
	rma1_set_mac(dev, dev->eth_mac);
	write32(GETHER_RFLR, 0x0800);
	write32(GETHER_PIPR, 0x00);
	write32(GETHER_APR, 0x4);
	write32(GETHER_MPR, 0x6);
	write32(GETHER_TPAUSER, 0xffff);

	write32(GETHER_ECMR, 0x60);

	mdio_read(dev, 0x00, 31, &val);
	if (val & 0x8)
		write32(GETHER_GECMR, 0x4); /* 100BASE */
	else
		write32(GETHER_GECMR, 0x0); /* 10BASE */

	if (val & 0x10)
		write32(GETHER_ECMR, 0x62); /* Full Duplex, rx/tx enable */
	else
		write32(GETHER_ECMR, 0x60); /* Half Duplex, rx/tx enable */

	write32(GETHER_ECMR, read32(GETHER_ECMR) | 0x20001);
	write32(GETHER_EDRRR, 1);

	return 0;
}

static int rma1_eth_shutdown(struct net_device *dev)
{
	trace();

	return 0;
}

struct net_device rma1_eth = {
	.init = rma1_eth_init,
	.exit = rma1_eth_shutdown,

	.set_mac = rma1_set_mac,
	.enable_phy = rma1_enable_phy,
	.disable_phy = rma1_disable_phy,
	.send = rma1_send,
	.recv = rma1_recv,
};
